// Vid80Dev.cpp: implementation of the CVid80Dev class.
// Copyright (C) HP-IL engine by J-F Garnier
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "ILVideo80.h"
#include "Vid80Dev.h"

// address in display memory
#define CHR_AT(x,y)		(m_pbyBuffer[(m_nBufLine + (y)) * m_nVisCols + (x)])

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CVid80Dev::CVid80Dev(CStaticTerm& wndScreen,INT nCols,INT nRows)
	: m_wndScreen(wndScreen),m_nVisCols(nCols),m_nVisRows(nRows),CIldevBase(true)
{
	// HP-IL data and variables
	m_byAID = 50;							// Accessory ID = video
	m_byDEFADDR = 31;						// default address after AAU
	m_strDID = _T("MC00701A\r\n");			// Device ID

	// video data and variables
	m_nBufRows = 48;						// no. of buffer rows
	m_nBufLine = 0;							// buffer line of screen view

	m_xCurPos = 0;							// x cursor position
	m_yCurPos = 0;							// y cursor position

	m_nEsc = 0; 							// no escape sequence
	m_bAutoCrLf = false;					// reset automatic CR LF flag

	m_bPal = false;							// NTSC mode
	m_bMonitorMode = false;					// monitor mode disabled
	m_bInsertMode = false;					// disable insert character mode

	// video buffer
	ASSERT(m_nVisCols <= 80);
	VERIFY(m_pbyBuffer = static_cast<LPBYTE>(new BYTE[80 * m_nBufRows]));
	FillMemory(m_pbyBuffer,80 * m_nBufRows,' ');

	// init auto CR LF line indicator
	VERIFY(m_pbyAutoCrLf = static_cast<LPBYTE>(new BYTE[m_nBufRows]));
	ZeroMemory(m_pbyAutoCrLf,m_nBufRows);

	m_wndScreen.SetAsciiMode();				// ASCII character set
	ASSERT(m_pbyBuffer);					// buffer must be allocated before calling Resize()
	Resize(m_nVisCols,m_nVisRows);			// set initial size of terminal window

	SyncScreen();							// synchronize screen buffer

	m_bCursorShow = true;					// show cursor
	m_eCurState = CUR_REPLACE;				// cursor is replace cursur
	SetCursorPos(0,0);						// set cursor
}

CVid80Dev::~CVid80Dev()
{
	delete[] m_pbyBuffer;					// video buffer
	delete[] m_pbyAutoCrLf;					// auto CR LF line indicator
}

//
// HP-IL status (2 bytes)
//
BYTE CVid80Dev::GetILStatus(WORD& wStatus)
{
	wStatus =  (0x80                                    << 8); // 'OK'
	wStatus |= (0                                       << 7);
	wStatus |= (0                                       << 6); // RAM U11 selftest (0=Good, 1=Bad)
	wStatus |= (0                                       << 5); // RAM U10 selftest (0=Good, 1=Bad)
	wStatus |= (((m_eCurState == CUR_INSERT)   ? 1 : 0) << 4); // Insert/Replace mode (0=Replace, 1=Insert)
	wStatus |= (((!m_bCursorShow)              ? 1 : 0) << 3); // Cursor (0=displayed, 1=not displayed)
	wStatus |= (((m_bPal)                      ? 1 : 0) << 2); // Video output (0=NTSC, 1=PAL)
	wStatus |= (((!m_wndScreen.GetAsciiMode()) ? 1 : 0) << 1); // Character set (0=ASCII/INV, 1=Roman-8)
	wStatus |= (((m_nVisCols == 40)            ? 1 : 0) << 0); // Screen width (0=80 col, 1=40 col)
	return 2;								// no. of status bytes
}

//
// synchronize screen with display memory buffer
//
VOID CVid80Dev::SyncScreen()
{
	LPTSTR lpszText;
	INT    nAddr,nBuflen;

	nAddr = m_nBufLine * m_nVisCols;		// begin inside display memory

	// buffer length
	nBuflen = (m_nVisCols + 2) * m_nVisRows + 1;

	// allocate output string buffer
	VERIFY(lpszText = static_cast<LPTSTR>(new TCHAR[nBuflen]));
	if (lpszText != NULL)					// buffer allocated
	{
		INT nLinAddr = 0;
		for (INT y = 0; y < m_nVisRows; ++y)
		{
			for (INT x = 0; x < m_nVisCols; ++x)
			{
				ASSERT(nLinAddr < nBuflen);
				ASSERT(nAddr < (m_nVisCols * m_nBufRows));
				lpszText[nLinAddr++] = static_cast<TCHAR>(m_pbyBuffer[nAddr++]);
				nAddr %= m_nVisCols * m_nBufRows; // address wrap around
			}
			lpszText[nLinAddr++] = _T('\r');
			lpszText[nLinAddr++] = _T('\n');
		}
		ASSERT(nLinAddr < nBuflen);
		lpszText[nLinAddr] = 0;				// EOS

		ASSERT(nLinAddr + 1 == nBuflen);	// text buffer has the expected length
		m_wndScreen.SetBufferText(lpszText);
		delete[] lpszText;
	}
	return;
}

//
// resize display and locate it inside border
//
VOID CVid80Dev::Resize(INT cx, INT cy)
{
	m_nVisCols = cx;						// no. of visible columns
	m_nVisRows = cy;						// no. of visible rows

	// new visible size
	ASSERT(m_wndScreen.IsKindOf(RUNTIME_CLASS(CStaticTerm)));
	m_wndScreen.SetSize(m_nVisCols,m_nVisRows);

	// clear display buffer
	FillMemory(m_pbyBuffer,80 * m_nBufRows,' ');
	ZeroMemory(m_pbyAutoCrLf,m_nBufRows);
	return;
}

//
// set new cursor position
//
VOID CVid80Dev::SetCursorPos(INT x, INT y)
{
	ASSERT (x >= 0 && x < m_nVisCols && y >= 0 && y < m_nVisRows);

	// set cursor state
	m_wndScreen.SetCursor(x,y,m_bCursorShow ? m_eCurState : CUR_OFF);

	m_xCurPos = x;
	m_yCurPos = y;
	return;
}

//
// move cursor left
//
VOID CVid80Dev::CursorLeft()
{
	// nove cursor left
	if (m_xCurPos-- == 0)					// already at first column
	{
		m_xCurPos = m_nVisCols - 1; 		// goto last column

		// move cursor up
		if (m_yCurPos-- == 0)				// already at first row
		{
			m_xCurPos = m_yCurPos = 0;		// set cursur to (0,0)
		}
	}
	return;
}

//
// move cursor right
//
VOID CVid80Dev::CursorRight()
{
	if (m_xCurPos++ == m_nVisCols - 1)
	{
		m_xCurPos = 0;
		if (m_yCurPos++ == m_nVisRows - 1)
			m_yCurPos = 0;
	}
	return;
}

//
// insert blank line
//
VOID CVid80Dev::InsertLine()
{
	m_xCurPos = 0;							// set cursor to column 0

	// move data from current buffer line one line down until end of buffer
	ASSERT(m_nBufRows > 2);
	LPBYTE pbyAct = &m_pbyBuffer[(m_nBufRows - 1) * m_nVisCols];
	for (INT i = m_nBufRows - 1; i > m_nBufLine + m_yCurPos; --i)
	{
		LPBYTE pbyDest = pbyAct;
		pbyAct -= m_nVisCols;

		// copy one line
		CopyMemory(pbyDest,pbyAct,m_nVisCols);
		m_pbyAutoCrLf[i] = m_pbyAutoCrLf[i - 1];
	}

	FillMemory(pbyAct,m_nVisCols,' ');		// fill actual line with spaces
	m_pbyAutoCrLf[m_nBufLine] = 0;
	SyncScreen();							// synchronize screen buffer
	return;
}

//
// delete current line
//
VOID CVid80Dev::DeleteLine()
{
	INT nLine = m_nBufLine + m_yCurPos;

	// move data from next line to the actual line until end of buffer
	ASSERT (nLine < m_nBufRows);
	ASSERT (m_nBufRows >= (nLine + 1));
	LPBYTE pbyDest = &m_pbyBuffer[nLine       * m_nVisCols];
	LPBYTE pbySrc  = &m_pbyBuffer[(nLine + 1) * m_nVisCols];

	// no. of lines to copy
	INT nNoOfLines = m_nBufRows - 1 - nLine;

	// copy the lines
	CopyMemory(pbyDest,pbySrc,nNoOfLines * m_nVisCols);
	CopyMemory(&m_pbyAutoCrLf[m_nBufLine],&m_pbyAutoCrLf[m_nBufLine + 1],nNoOfLines);
	m_pbyAutoCrLf[m_nBufRows - 1] = 0;

	// fill the last line with spaces
	FillMemory(&m_pbyBuffer[(m_nBufRows - 1) * m_nVisCols],m_nVisCols,' ');
	SyncScreen();							// synchronize screen buffer
	return;
}

//
// insert character at cursor position and move the rest of input line right
//
VOID CVid80Dev::InsertChar(BYTE byData)
{
	INT i,x,y,xc,yc,yi;

	bool bInsLine = false;

	x = xc = m_xCurPos;
	y = yc = m_yCurPos;

	// goto last input line
	while (m_pbyAutoCrLf[m_nBufLine + y] != 0)
	{
		x = 0;								// line begin with cusor position 0
		++y;								// next line
	}

	for (i = m_nVisCols; i > x; --i)		// goto end of line
	{
		if (CHR_AT(i - 1,y) != ' ')
			break;
	}
	x = i;									// x = position of last space in line

	if (x == m_nVisCols)					// no spaces left in last input line
	{
		x = 0;
		m_pbyAutoCrLf[m_nBufLine + y] = 1;
		++y;

		bInsLine = true;					// insert new line at (0,yi)
		yi = y;
	}

	if (++m_xCurPos == m_nVisCols)			// increment cursor position
	{
		m_xCurPos = 0;
		m_pbyAutoCrLf[m_nBufLine + m_yCurPos] = 1;
		++m_yCurPos;

		if (m_yCurPos > y)					// cursor in new line
		{
			bInsLine = true;				// insert new line at (0,yi)
			yi = m_yCurPos;
		}
	}

	// new line at end of visible screen necessary?
	if (   y         == m_nVisRows			// no spaces left
		|| m_yCurPos == m_nVisRows)			// cursor out of visible area
	{
		// end of buffer
		if (m_nBufLine + m_nVisRows == m_nBufRows)
		{
			// move buffer content one line up to free the last line
			CopyMemory(m_pbyBuffer,m_pbyBuffer + m_nVisCols,(m_nBufRows - 1) * m_nVisCols);
			CopyMemory(m_pbyAutoCrLf,m_pbyAutoCrLf + 1,m_nBufRows - 1);

			// fill the last line with spaces
			FillMemory(&m_pbyBuffer[(m_nBufRows - 1) * m_nVisCols],m_nVisCols,' ');
			m_pbyAutoCrLf[m_nBufRows - 1] = 0;
			bInsLine = false;
		}
		else
		{
			++m_nBufLine;					// move visible area one line down
		}

		// adjust cursors to new buffer end
		--y;
		--yc;
		--yi;
		--m_yCurPos;
	}

	if (bInsLine)							// insert new line at yi
	{
		yi += m_nBufLine;					// absolute line no. in buffer

		LPBYTE pbySrc  = &m_pbyBuffer[yi * m_nVisCols];
		LPBYTE pbyDest = pbySrc + m_nVisCols;

		INT i = m_nBufRows - 1 - yi;		// no. of lines to move

		// free line at yi
		MoveMemory(pbyDest,pbySrc,i * m_nVisCols);
		MoveMemory(&m_pbyAutoCrLf[yi],&m_pbyAutoCrLf[yi+1],i);
		m_pbyAutoCrLf[m_nBufRows - 1] = 0;

		FillMemory(pbySrc,m_nVisCols,' ');	// clear new line
		m_pbyAutoCrLf[yi] = 0;
	}

	LPBYTE pbyStart = &CHR_AT(xc,yc);		// linear position of cursor in buffer

	// move data from original cursor position one character right
	UINT uDataLength = static_cast<UINT>(&CHR_AT(x,y) - pbyStart);
	if (uDataLength > 0)
	{
		MoveMemory(pbyStart + 1,pbyStart,uDataLength);
	}
	*pbyStart = byData;						// write character
	SyncScreen();							// synchronize screen buffer
	return;
}

//
// delete character at cursor position and move the rest of input line left
//
VOID CVid80Dev::DeleteChar()
{
	INT yCurPos = m_nBufLine + m_yCurPos;
	INT nSize = m_nVisCols - 1 - m_xCurPos;
	INT nEmptyLines = 0;

	// check for following lines
	while (m_pbyAutoCrLf[yCurPos] != 0 && ++yCurPos < m_nBufRows)
	{
		nSize += m_nVisCols;
	}

	// delete character at cursor position and move rest of line left
	LPBYTE pbyData = &CHR_AT(m_xCurPos,m_yCurPos);
	CopyMemory(pbyData,pbyData + 1,nSize);

	pbyData[nSize] = ' ';					// clear last character in line

	if (nSize >= m_nVisCols)				// input line is longer than screen line
	{
		INT nPos;

		// check for last character in input line
		for (nPos = nSize; nPos > 0; --nPos)
		{
			if (pbyData[nPos] != ' ')		// found non space character
				break;
		}

		// calculate number of empty lines
		nEmptyLines = (nSize - nPos) / m_nVisCols;
	}

	if (nEmptyLines > 0)					// have to clear screen lines
	{
		INT i;

		// clear the empty lines
		LPBYTE pbySrc  = &m_pbyBuffer[++yCurPos * m_nVisCols];
		LPBYTE pbyDest = pbySrc - nEmptyLines * m_nVisCols;

		i = m_nBufRows - yCurPos;			// no. of lines to copy
		if (i > 0)							// have to copy lines
		{
			ASSERT(pbyDest < pbySrc);
			CopyMemory(pbyDest,pbySrc,i * m_nVisCols);
			CopyMemory(&m_pbyAutoCrLf[yCurPos-nEmptyLines],&m_pbyAutoCrLf[yCurPos],i);
			m_pbyAutoCrLf[yCurPos-nEmptyLines-1] = 0;
		}

		// fill buffer with spaces at the end
		i = m_nBufRows - nEmptyLines;		// start buffer line to fill
		FillMemory(&m_pbyBuffer[i * m_nVisCols],nEmptyLines * m_nVisCols,' ');
		ZeroMemory(&m_pbyAutoCrLf[i],nEmptyLines);
	}
	SyncScreen();							// synchronize screen buffer
	return;
}

//
// eval line feed
//
VOID CVid80Dev::EvalLF()
{
	if (++m_yCurPos == m_nVisRows)			// end of screen
	{
		m_yCurPos = m_nVisRows - 1;			// set cursor to last line

		// not end of buffer
		if (m_nBufLine + m_nVisRows < m_nBufRows)
		{
			++m_nBufLine;					// scroll up buffer
		}
		else // end of buffer
		{
			ASSERT(m_nBufLine + m_nVisRows == m_nBufRows);

			// delete 1st buffer line and move the rest lines one line up
			CopyMemory(m_pbyBuffer,m_pbyBuffer + m_nVisCols,(m_nBufRows - 1) * m_nVisCols);
			CopyMemory(&m_pbyAutoCrLf[0],&m_pbyAutoCrLf[1],(m_nBufRows - 1));
			ASSERT(m_nBufLine + m_yCurPos == m_nBufRows - 1);
		}

		// clear last line of window
		FillMemory(&m_pbyBuffer[(m_nBufLine + m_yCurPos) * m_nVisCols],m_nVisCols,' ');
		m_pbyAutoCrLf[m_nBufLine + m_yCurPos] = 0;

		SyncScreen();						// synchronize screen buffer
	}
	return;
}

//
// Soft reset
// empty display & screen buffer and show replace cursor at (0,0)
//
VOID CVid80Dev::SoftReset()
{
	// empty display buffer
	FillMemory(m_pbyBuffer,80 * m_nBufRows,' ');
	ZeroMemory(m_pbyAutoCrLf,m_nBufRows);
	m_nBufLine = 0;							// view display buffer from top
	SyncScreen();							// synchronize screen buffer
	m_bCursorShow = true;					// show
	m_eCurState = CUR_REPLACE;				// replace
	SetCursorPos(0,0);						// cursor at (0,0)
	return;
}

//
// Hard reset
// empty display & screen buffer and show replace cursor at (0,0)
// set all soft switches to their default position
//
VOID CVid80Dev::ClearDevice()
{
	// set all soft switches to their default position
	m_bPal = false;							// NTSC mode
	m_wndScreen.SetAsciiMode();				// ASCII character set
	m_bMonitorMode = false;					// monitor mode disabled
	m_bInsertMode = false;					// disable insert character mode
	Resize(80,24);							// set initial size of terminal window

	SoftReset();							// then do the soft reset
	return;
}

//
// received data byte in listener mode
//
VOID CVid80Dev::InData(WORD wFrame)
{
	INT i,j;

	BYTE byData = static_cast<BYTE>(wFrame);

	// this is the virtual loop
	TRACE(_T("ILVideo Data = %02X\n"),byData);
//	TRACE(_T("ILVideo Data = %02X '%c'\n"),byData,_istprint(byData) ? byData : _T('.'));

	if (m_nEsc > 0 && !m_bMonitorMode) 		// escape sequence
	{
		--m_nEsc;							// handled ESC sequence

		switch (m_nEsc)
		{
		case 0:								// single ESC
			switch (byData)
			{
			case '>': // Cursor On
				m_bCursorShow = true;
				break;
			case '<': // Cursor Off
				m_bCursorShow = false;
				break;
			case 'A': // Cursor Up
				if (m_yCurPos-- == 0)		// move cursor up
					m_yCurPos = 0;
				break;
			case 'B': // Cursor Down
				if (m_yCurPos++ == m_nVisRows - 1)
					m_yCurPos = m_nVisRows - 1;
				break;
			case 'C': // Cursor Right
				CursorRight();
				break;
			case 'D': // Cursor Left
				CursorLeft();
				break;
			case 'E': // Soft reset
				SoftReset();
				break;
			case 'H': // Cursor Home
				m_xCurPos = m_yCurPos = 0;
				break;
			case 'J': // Clear screen memory from cursor to end of page
				i = m_xCurPos;				// first clear until end of current line

				// then clear the lines after current position
				for (j = m_yCurPos; j < m_nVisRows; ++j)
				{
					// clear automatic Crlf marker
					m_pbyAutoCrLf[m_nBufLine + j] = 0;

					for (; i < m_nVisCols; ++i)
					{
						CHR_AT(i,j) = ' ';
					}
					i = 0;					// clear all other lines from beginning
				}
				SyncScreen();				// synchronize screen buffer
				break;
			case 'K': // Clear screen memory from cursor to end of line
				// clear until end of current line
				for (i = m_xCurPos; i < m_nVisCols; ++i)
				{
					CHR_AT(i,m_yCurPos) = ' ';
				}

				i = m_nBufLine + m_yCurPos;	// cursor line in buffer

				// clear automatic Crlf marker lines
				for (j = i; m_pbyAutoCrLf[j] != 0; ++j)
				{
					m_pbyAutoCrLf[j] = 0;	// clear automatic Crlf marker
				}

				// i = cursor line in buffer
				// j = last line in buffer to purge
				if (i != j)					// have to purge lines
				{
					INT nFill,nCopy;

					nFill = j - i;			// lines to fill at end

					LPBYTE pbyDest = &m_pbyBuffer[++i * m_nVisCols];
					LPBYTE pbySrc  = &m_pbyBuffer[++j * m_nVisCols];

					nCopy = m_nBufRows - j;	// no. of lines to copy
					if (nCopy > 0)			// have to copy lines
					{
						ASSERT(pbyDest < pbySrc);
						CopyMemory(pbyDest,pbySrc,nCopy * m_nVisCols);
						CopyMemory(&m_pbyAutoCrLf[i],&m_pbyAutoCrLf[j],nCopy);
					}

					// fill buffer with spaces at the end
					i = m_nBufRows - nFill;	// start buffer line to fill
					FillMemory(&m_pbyBuffer[i * m_nVisCols],nFill * m_nVisCols,' ');
					ZeroMemory(&m_pbyAutoCrLf[i],nFill);
				}
				SyncScreen();				// synchronize screen buffer
				break;
			case 'L': // Insert line
				InsertLine();
				break;
			case 'M': // Delete line
				DeleteLine();
				break;
			case 'N': // Enable insert character mode
				m_eCurState = CUR_INSERT;
				m_bInsertMode = true;		// enable insert character mode
				break;
			case 'O': // Delete character
				DeleteChar();
				break;
			case 'Q': // Insert Cursor
				m_eCurState = CUR_INSERT;
				break;
			case 'R': // Replace Cursor
				m_eCurState = CUR_REPLACE;
				m_bInsertMode = false;		// disable insert character mode
				break;
			case 'S': // Scroll Up (move window down)
				if (m_nBufLine + m_nVisRows < m_nBufRows)
				{
					++m_nBufLine;			// scroll up not circular
					SyncScreen();			// synchronize screen buffer
				}
				break;
			case 'T': // Scroll Down (move window up)
				if (m_nBufLine > 0)
				{
					--m_nBufLine;			// scroll down not circular
					SyncScreen();			// synchronize screen buffer
				}
				break;
			case 'Y': // Enable Monitor Mode
				m_bMonitorMode = true;
				break;
			case 'Z': // Disable Monitor Mode
				m_bMonitorMode = false;
				break;
			case '[': // Display 80 columns by 24 lines
				m_xCurPos = m_yCurPos = 0;	// cursor position 0,0
				m_nBufLine = 0;				// buffer position 0,0
				Resize(80,24);				// new display window size
				SyncScreen();				// synchronize screen buffer
				break;
			case ']': // Display 40 columns by 20 lines
				m_xCurPos = m_yCurPos = 0;	// cursor position 0,0
				m_nBufLine = 0;				// buffer position 0,0
				Resize(40,20);				// new display window size
				SyncScreen();				// synchronize screen buffer
				break;
			case '/': // Switch to NTSC mode
				m_bPal = false;
				break;
			case '=': // Switch to PAL mode
				m_bPal = true;
				break;
			case 'e': // Hard reset
				ClearDevice();
				break;
			case 'h': // Home cursor (screen memory)
				m_xCurPos = m_yCurPos = 0;	// cursor position 0,0
				m_nBufLine = 0;				// buffer position 0,0
				SyncScreen();				// synchronize screen buffer
				break;
			case 'j': // Use Roman-8 character set
				m_wndScreen.SetAsciiMode(false);
				break;
			case 'k': // Use standard character set
				m_wndScreen.SetAsciiMode(true);
				break;
			case 'z': // performs an extensive selft test on the internal screen RAM
				break;
			case '%': // Cursor To Address
				m_nEsc = 3; 				// still need two arguments, fetch column next
				break;
			default: // illegal ESC sequence
				break;						// ignore sequence
			}
			break;

		case 2: // Cursor To Address, m column argument
			m_nEscCol = byData;				// save column argument
			m_nEsc = 2; 					// still need one argument, fetch row next
			break;

		case 1: // Cursor To Address, n row argument
			// column and row argument inside screen
			if (m_nEscCol < m_nVisCols && byData < m_nVisRows)
			{
				m_xCurPos = m_nEscCol;
				m_yCurPos = byData;
			}
			m_nEsc = 0; 					// handled ESC sequence
			break;

		default:
			ASSERT(false);
		}

		m_bAutoCrLf = false;				// reset automatic CR LF flag
	}
	else
	{
		do
		{
			// get current character set
			bool bAscii = m_wndScreen.GetAsciiMode();

			// displayable character?
			if (   ( bAscii && ((byData & 0x7F) >= 32 && (byData & 0x7F) <= 127))
				|| (!bAscii && (byData >= 32 && byData <= 255))
				|| (m_bMonitorMode && byData > 0)
			   )
			{
				m_bAutoCrLf = false;		// reset automatic CR LF flag

				if (m_bInsertMode)			// insert character mode activated?
				{
					InsertChar(byData);		// move rest of line to right and show character
				}
				else						// insert character mode not activated?
				{
					CHR_AT(m_xCurPos,m_yCurPos) = byData;
					SyncScreen();			// synchronize screen buffer

					// end of column?
					if (++m_xCurPos == m_nVisCols)
					{
						// set automatic CrLf marker for current line
						m_pbyAutoCrLf[m_nBufLine + m_yCurPos] = 1;

						m_xCurPos = 0;		// eval CR
						EvalLF();			// eval LF
						m_bAutoCrLf = true; // evaluated automatic CR LF
					}
				}

				// check for ESC 'Z' in Monitor Mode
				if (m_nEsc > 0) 			// escape sequence
				{
					--m_nEsc;				// handled ESC sequence

					if (byData == 'Z')		// Disable Monitor Mode
					{
						m_bMonitorMode = false;
					}
				}
			}

			// BS and not in monitor mode
			if (byData == 8 && !m_bMonitorMode)
			{
				CursorLeft();				// move cursor left
				m_bAutoCrLf = false;		// reset automatic CR LF flag
				break;
			}
			// LF and not in monitor mode
			if (byData == 10 && !m_bMonitorMode)
			{
				if (!m_bAutoCrLf)			// not evaluated automatic CR LF
				{
					EvalLF();				// eval LF
				}
				else						// already evaluated automatic CR LF
				{
					ASSERT(m_nBufLine + m_yCurPos > 1);
				}
				// delete possible automatic CrLf marker of prior line
				m_pbyAutoCrLf[m_nBufLine + m_yCurPos - 1] = 0;
				m_bAutoCrLf = false;		// reset automatic CR LF flag
				break;
			}
			// CR and not in monitor mode
			if (byData == 13 && !m_bMonitorMode)
			{
				m_xCurPos = 0;				// cursor to column 0
				break;
			}
			// ESC
			if (byData == 27)
			{
				m_nEsc = 1;					// escape sequence
				break;
			}
		}
		while (false);
	}

	SetCursorPos(m_xCurPos,m_yCurPos);		// update cursor
}
